/**********************************************************************************

  DELTAGAMMAINC Fast and Accurate Evaluation of a Generalized Incomplete Gamma
  Function. Copyright (C) 2016 Remy Abergel (remy.abergel AT gmail.com), Lionel
  Moisan (Lionel.Moisan AT parisdescartes.fr).

  This file is a part of the DELTAGAMMAINC software, dedicated to the
  computation of a generalized incomplete gammafunction. See the Companion paper
  for a complete description of the algorithm.

  ``Fast and accurate evaluation of a generalized incomplete gamma function''
  (Rémy Abergel, Lionel Moisan), preprint MAP5 nº2016-14.

  This program is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation, either version 3 of the License, or (at your option) any later
  version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.

  You should have received a copy of the GNU General Public License along with
  this program.  If not, see <http://www.gnu.org/licenses/>.

**********************************************************************************/

/*
 * Description: compute gamma(mu,p,x) or Gamma(mu,p,x), where
 *
 *    gamma(mu,p,x) = integral over [0,x] of s^(p-1) * exp(-mu*s) ds
 *    Gamma(mu,p,x) = integral over [x,inf] of s^(p-1) * exp(-mu*s) ds
 *
 * using specific computation method (continued fractions, recursive integration
 * by parts, series expansions).
 *
 * Compilation (with gcc): gcc -O3 gammainc.c gammainc_tools.c -lm -o gammainc_tools
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

static void display_usage()
{
  printf("Usage: gammainc_tools [--help] [-t t] [-u] [-s] [-c] [-i] x mu p\n\n");
  printf("   --help : display help\n");
  printf("   -t t   : recompute the integral until elapsed time >= t (default t=0.1s) \n");
  printf("   -u     : compute Gamma(mu,p,x) using a continued fraction (mu > 0)\n");
  printf("   -s     : compute gamma(mu,p,x) using a series expansion\n");
  printf("   -c     : compute gamma(mu,p,x) using a continued fraction\n");
  printf("   -i     : compute gamma(mu,p,x) using a recursive integration by parts (mu < 0)\n");
  printf("   x      : (double) a nonnegative, possibly infinite (inf) number\n");
  printf("   mu     : (double) a nonzero real number\n");
  printf("   p      : (double) a real number >=1 (must be integer if mu<0)\n\n");
  printf("Description:\n\n");
  printf("   Compute (rho,sigma) such as I = rho * exp(sigma), where\n");
  printf("   I = gamma(mu,p,x) or Gamma(mu,p,x), and,\n\n");
  printf("   gamma(mu,p,x) = integral over [0,x] of s^{p-1} * exp(-mu*s) ds,\n");
  printf("   Gamma(mu,p,x) = integral over [x,inf] of s^{p-1} * exp(-mu*s) ds.\n\n");
}

int main(int argc, char **argv)
{
  char *x_value = NULL,*mu_value = NULL,*p_value = NULL,*t_value = NULL,uflag=0,cflag=0,sflag=0,iflag=0;
  clock_t start,end;
  double rho,sigma,x,mu,p,t=0.1,elapsed_time,a,b,b0,l10;
  int err,i,N=0;

  /* parser */
  for(i=1;i<argc;i++) {
    if (strcmp(argv[i],"--help") == 0) {
      display_usage();
      return EXIT_SUCCESS;
    }
    else if (strcmp(argv[i],"-u") == 0) {
      uflag = 1;
    }
    else if (strcmp(argv[i],"-s") == 0) {
      sflag = 1;
    }
    else if (strcmp(argv[i],"-c") == 0) {
      cflag = 1;
    }
    else if (strcmp(argv[i],"-i") == 0) {
      iflag = 1;
    }
    else if (strcmp(argv[i],"-t") == 0) {
      if(i==argc-1) {
	printf("\nError: option -t requires an argument.\n");
	display_usage();
	return EXIT_FAILURE;
      }
      else {
	t_value = argv[i+1]; err = sscanf(t_value,"%lf",&t); i++;
	if(err != 1) {
	  printf("\nError: could not properly retrieve the argument of option -t.\n");
	  display_usage();
	  return EXIT_FAILURE;
	}
      }
    }
    else if(NULL == x_value) {
      x_value = argv[i];
      err = sscanf(x_value,"%lf",&x);
      if(err != 1) {
	printf("\nError: could not retrieve properly input argument 'x'.\n");
	display_usage();
	return EXIT_FAILURE;
      }
    }
    else if(NULL == mu_value) {
      mu_value = argv[i];
      err = sscanf(mu_value,"%lf",&mu);
      if(err != 1) {
	printf("\nError: could not retrieve properly input argument 'mu'.\n");
	display_usage();
	return EXIT_FAILURE;
      }
    }
    else if(NULL == p_value) {
      p_value = argv[i];
      err = sscanf(p_value,"%lf",&p);
      if(err != 1) {
	printf("\nError: could not retrieve properly input argument 'p'.\n");
	display_usage();
	return EXIT_FAILURE;
      }
    }
  }
  if ((NULL == x_value) || (NULL == mu_value) || (NULL == p_value)) {
    printf("\nError: invalid number of arguments.\n");
    display_usage();
    return EXIT_FAILURE;
  }

  /* consistency checks */
  if(uflag+sflag+cflag+iflag != 1) {
    printf("\nError: you must select exactly one option among (-u,-s,-c,-i)!\nType gammainc_tools --help to display help.\n\n");
    return EXIT_FAILURE;
  }
  if(x < 0) { printf("\nError: input 'x' must be nonnegative!\nType gammainc_tools --help to display help.\n\n"); return EXIT_FAILURE; }
  if(uflag && (mu < 0)) { printf("\nError: Gamma(mu,p,x) is indefinite when mu < 0!\nType gammainc_tools --help to display help.\n\n"); return EXIT_FAILURE; }
  if(iflag && (mu > 0)) { printf("\nError: the recursive integration by part cannot be used with mu > 0!\nType gammainc_tools --help to display help.\n\n"); return EXIT_FAILURE; }
  if(iflag && ((-mu*x <= p-1)||(-mu*x <= 1))) { printf("\nError: the recursive integration by part must be used with |mu*x| > max(1,p-1)!\nType gammainc_tools --help to display help.\n\n"); return EXIT_FAILURE; }
  if(p < 1) { printf("\nError: you must choose input 'p' such as (p >= 1)!\nType deltagammainc-smart --help to display help.\n\n"); return EXIT_FAILURE; }
  if((mu < 0) && (p != floor(p))) { printf("\nError: non integer values for input 'p' are not allowed when mu < 0!\nType deltagammainc-smart --help to display help.\n\n"); return EXIT_FAILURE; }
  if(mu == 0) { printf("\nError: input 'mu' must be nonzero!\nType gammainc_tools --help to display help.\n\n"); return EXIT_FAILURE; }
  if(isinf(x) && (mu < 0)) { printf("\nError: input 'x' cannot be infinite when input 'mu' is negative, because\n the integral gamma(mu,p,x) is indefinite for that setting!\n\nType gammainc_tools --help to display help.\n\n"); return EXIT_FAILURE; }

  /* kernel (compute lower or upper incomplete gamma function) */
  if(uflag) { // compute the upper incomplete gamma function (using a continued fraction)
    start = clock();
    do {
      Gammainc_cfrac(&rho,&sigma,mu,p,x);
      N++; end = clock();
    } while(((double) end-start)/CLOCKS_PER_SEC < t);
  }
  else if(sflag) { // compute the lower incomplete gamma function (using a series expansion)
    start = clock();
    do {
      gammainc_ser(&rho,&sigma,mu,p,x);
      N++; end = clock();
    } while(((double) end-start)/CLOCKS_PER_SEC < t);
  }
  else if(cflag) { // compute the lower incomplete gamma function (using a continued fraction)
    start = clock();
    do {
      gammainc_cfrac(&rho,&sigma,mu,p,x);
      N++; end = clock();
    } while(((double) end-start)/CLOCKS_PER_SEC < t);
  }
  else if(iflag) { // (only when mu < 0) compute the lower incomplete gamma function (using a recursive integration by parts)
    start = clock();
    do {
      gammainc_ibp(&rho,&sigma,mu,p,x);
      N++; end = clock();
    } while(((double) end-start)/CLOCKS_PER_SEC < t);
  }
  elapsed_time = ((double) end-start) / CLOCKS_PER_SEC;

  /*** printf results ***/
  if(uflag) printf("\nComputing I = integral over [x,inf] of s^{p-1} * exp(-mu*s) ds,\n");
  else printf("\nComputing I = integral over [0,x] of s^{p-1} * exp(-mu*s) ds,\n");
  printf("where (hexadecimal): x=%a, mu=%a, p=%a,\n",x,mu,p);
  printf("(in decimal approx): x=%.34g, mu=%.34g, p=%.34g.\n\n",x,mu,p);

  /* manage display with decomposition I = a*10^(b) where a in [1,10) and b integer */
  err = (rho < 0 || isnan(rho) || isinf(rho));
  if(err) printf("Output: I = error! Some numerical unstabilities occured, try\nanother computation method.\n\n");
  else if (rho==0 || (isinf(sigma) && sigma < 0)) printf("Output: I = %.17e\n\n",0.);
  else {
    a=rho*exp(sigma);
    //if(a>0. && !isinf(a)) printf("Output: I = %.17e\n\n",a);
    if(a>1e-308 && a<1e308) printf("Output: I = %.17e\n\n",a);
    else {
      l10 = log(10);
      b = sigma/l10;
      b0 = floor(b+log10(rho));
      a = rho*exp((b-b0)*l10);
      printf("Output: I = %.17fe%+03d\n\n",a,(int)b0);
    }
  }

  /* display the computed mantissa-exponent representation */
  printf("Representation with double numbers: I = rho*exp(sigma), where\n");
  printf("rho   = %.17e\n",rho);
  printf("sigma = %.17e\n\n",sigma);
  printf("computation time (nanosec): %.5g\n\n",1.e9*(elapsed_time/(double)N));

  return EXIT_SUCCESS;
}
